# ficc - File Integrity Command & Control - monitor Tripwire installations
# Copyright (C) 2003  Terry D. Ott
#
# Version 1.2 -- 23 November 2003
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package FICC;

use strict;
use vars qw( @ISA @EXPORT );

use Exporter;

@ISA = qw( Exporter );

@EXPORT = qw( getConfigFile readConfigFile readChecksumDB );


# Check the command line arguments to find the config file.  If no -c 
# argument is specified, look for the file in ./etc/ficc.conf
sub getConfigFile
{
   my $binaryDir = $_[ 0 ];

   my $cfgFile;

   if ( scalar( @ARGV ) > 0 )
   {
      if ( ( scalar( @ARGV ) != 2 ) || ( $ARGV[ 0 ] ne '-c' ) )
      {
         printHelp();
      }

      $cfgFile = $ARGV[ 1 ];
   }
   else
   {
      $cfgFile = "$binaryDir/../etc/ficc.conf";
   }

   return $cfgFile;
}


# Open and read configuration file
sub readConfigFile
{
   my $cfgFile = $_[ 0 ];
   my @validGlobalKeys = @{ $_[ 1 ] };
 
   my %hostInfo;
   my %globalConfigInfo;
   my $currLine;
   my $currHeader;
   my $cfgFileLine = 0;
   my @validKeys = ( 'email', 'host', 'login', 'cfgfile',
      'twexec', 'db', 'twopts', 'mail_subj', 'mail_if_unchanged',
      'quick_check', 'remote_md5_exec' );


   if ( ! open( CFGFILE, "< $cfgFile" ) )
   {
      print("Fatal: could not open configuration file $cfgFile. Try using " .
         "the -c option\n");

      printHelp();
   }

   while ( $currLine = <CFGFILE> )
   {
      $cfgFileLine++;

      chomp( $currLine );

      # Ignore comments and all-whitespace lines
      if ( ( $currLine =~ /^\s*#/ ) || ( $currLine =~ /^\s*$/ ) )
      {
         next;
      }

      # Is this line a header for a new section? 
      elsif ( $currLine =~ /\[.*\]/ )
      {
         # Remove all whitespace
         $currLine =~ s/\s//g;
         
         if ( $currLine =~ /^\s*$/ )
         {
            print( "Invalid system name at line $cfgFileLine of " .
               "$cfgFile\n" );
            exit -1;
         }

         $currHeader = $currLine;
         $currHeader =~ s/\[(.*)\]/$1/;

      }

      # Is this line a key/value pair?
      elsif( $currLine =~ /=/ )
      {
         my $key;
         my $value;

         ($key, $value) = split( /=/, $currLine );

         # Clean up any obnoxious whitespace
         $key   =~ s/^\s*(\S.*\S)\s*$/$1/;
         $value =~ s/^\s*(\S.*\S)\s*$/$1/;

         # Is this value in the global_options section?
         if ( $currHeader eq 'global_options' )
         {
            # Check the key value against valid global options
            if ( ! grep( /$key/, @validGlobalKeys ) )
            {
               print( "Fatal error on line $cfgFileLine of $cfgFile:\n" .
                  "Invalid global option: \"$key\"\n" );

               exit -1;
            }

            # Did we previously store a value under this key?
            if ( $globalConfigInfo{ $key } )
            {
               print( "Fatal error on line $cfgFileLine of $cfgFile: " .
                  "redeclaration for global option \"$key\"\n" );

               exit -1;
            }

            # Does the key have a value?
            if ( $value =~ /^\s*$/ )
            {
               print( "Fatal error on line $cfgFileLine of $cfgFile: " .
                  "value for required key \"$key\" is blank\n" );

               exit -1;
            }

            # Store the information in the global info hash
            $globalConfigInfo{ $key } = $value;
         }

         # Key/value pair for normal host
         elsif ( grep( /$key/, @validKeys ) )
         {
            # Get the key/value hash for this host entry
            my %keyValuePairs;

            if( $hostInfo{ $currHeader } )
            {
               %keyValuePairs = %{ $hostInfo{ $currHeader } };
            }

            # Did we already have a value for this key in this host?
            if ( $keyValuePairs{ $key } )
            {
               print("Fatal error on line $cfgFileLine of $cfgFile: " .
                  "redeclaration for key \"$key\" in section " .
                  "\"$currHeader\"\n" );

               exit -1;
            }

            # Insert our data into the hash
            $keyValuePairs{ $key } = $value;

            # Store the new info
            $hostInfo{ $currHeader } = \%keyValuePairs;
         }
         else
         {
            print "Fatal error: invalid key at line $cfgFileLine " .
               "of $cfgFile: \"$key\"\n";

            exit -1;
         }
      }

      # We must have a poorly constructed line, puke out
      else
      {
         print "Fatal error: invalid configuration at line $cfgFileLine " .
            "of $cfgFile\n";
         exit -1;
      }
   }

   close( CFGFILE );

   # Verify all required keys are there for all hosts
   checkConfigInfo( \%globalConfigInfo, \%hostInfo, \@validGlobalKeys );

   return ( \%globalConfigInfo, \%hostInfo );
}

sub checkConfigInfo
{
   my %globalConfigInfo = %{ $_[ 0 ] };
   my %hostConfigInfo = %{ $_[ 1 ] };
   my @validGlobalKeys = @{ $_[ 2 ] };

   my $currKey;

   # Verify global config info
   foreach $currKey ( @validGlobalKeys )
   {
      if ( ! grep( /$currKey/, keys( %globalConfigInfo ) ) )
      {
         print( 'Fatal error: configuration file does not contain required ' .
            " value \"$currKey\"\n" );

         exit -1;
      }
   }

   my @requiredCfgInfo = ( 'email', 'host', 'cfgfile', 'twexec',
      'db' );


   foreach $currKey ( keys( %hostConfigInfo ) )
   {
      my $checkKey;

      # Check for config file that existed but contained no key/value pairs
      if ( ! $hostConfigInfo{ $currKey } )
      {
         print( "Fatal error: configuration file contains no information " .
            "for host $currKey\n" );

         exit -1;
      }

      my %hostDetails = %{ $hostConfigInfo{ $currKey } };

      foreach $checkKey ( @requiredCfgInfo )
      {

         # Verify each host has all the required keys
         if ( ! grep( /$checkKey/, keys( %hostDetails ) ) )
         {
            print("Fatal error: configuration file does not contain " .
               "required value \"$checkKey\" for system \"$currKey\"\n" );

            exit( -1 );
         }
         else
         {
            # Okay, the key exists, does it have a value?
            if ( ! $hostDetails{ $checkKey } )
            {
               print("Fatal error: configuration file has required entry " .
                  "\"$checkKey\", but there is no value for that entry\n" );

               exit( -1 );
            }
         }
      }

      # Did the user have quick_check set, but left off remote_md5_exec?
      if ( ( $hostDetails{ 'quick_check' } eq 'yes' ) &&
         ( ! $hostDetails{ 'remote_md5_exec' } ) )
      {
         print('Fatal error: configuration file has quick_check set for ' .
            "host $currKey, but remote_md5_exec is not set" );
         exit( -1 );
      }
   }
}

sub readChecksumDB
{
   my $dbDir = $_[ 0 ];
   my %hostConfigInfo = %{ $_[ 1 ] };
   my %checksumData;
   my @fileList;

   my @systems = keys( %hostConfigInfo );

   my $currSystem;

   foreach $currSystem ( @systems )
   {
      if ( ! -d "$dbDir/$currSystem" )
      {
         print( "Fatal error: System \"$currSystem\" existed in the " .
            'configuration file but does not have a database directory ' .
            "in $dbDir\n" ); 

         exit -1;
      }

      my %hostInfo = %{ $hostConfigInfo{ $currSystem } };

      # Read checksums for important files (config, executable, database)

      my %hostChecksums;
      if ( $checksumData{ $currSystem } )
      {
         %hostChecksums = %{ $checksumData{ $currSystem } };
      }

      my @fileList = ( 'cfgfile', 'twexec', 'db' );
      if ( $hostInfo{ 'quick_check' } eq 'yes' )
      {
         push( @fileList, 'remote_md5_exec' );
      }

      my $hashKey;
      foreach $hashKey ( @fileList )
      {
         # Get fully-qualified filename 
         my $fileName = $hostInfo{ $hashKey };
         
         # Remove everything up to and including the last forward slash
         $fileName =~ s/^.*\///;

         # Prepend the db directory and system, then append the 
         # .cksum extension
         $fileName = "$dbDir/$currSystem/$fileName.cksum";

         if ( ! -f $fileName )
         {
            print( "Fatal error: checksum file \"$fileName\" does not " .
               "exist for system \"$currSystem\"\n" );

            exit -1;
         }

         if ( ! open( CKSUM, $fileName ) )
         {
            print("Fatal error: Error when opening checksum file " .
               "$fileName\n");

            exit -1;
         }

         # Read checksum info
         my $readLine = <CKSUM>;
         chomp( $readLine );

         close( CKSUM );

         # Update hash for this host
         $hostChecksums{ $hashKey } = $readLine;

         # Store hash reference in top level checksum hash
         $checksumData{ $currSystem } = \%hostChecksums;

      }
   }

   return \%checksumData;
}

1;
